(* ==================================================================== *)
(*									*)
(*  State Module for the Gardens Point Component Pascal Compiler.	*)
(*	Copyright (c) John Gough 1999, 2000.				*)
(*									*)
(*  Note that since this module is likely to be imported by most other  *)
(*  modules, it is important to ensure that it does not import others,  *)
(*  to avoid import cycles. 						*)
(*									*)
(* ==================================================================== *)

MODULE CompState;

  IMPORT 
	GPCPcopyright,
	RTS,
	Error,
	GPText,
	Symbols,
	IdDesc,
	Console, 
	CPascalS,
	NameHash,
	FileNames,
	CPascalErrors;

  CONST
     (*	VERSION    = "0.1 of 26 December 1999"; *)
     (*	VERSION    = "0.2 of 01 March 2000"; *)
     (*	VERSION    = "0.3 of 22 April 2000"; *)
     (*	VERSION    = "0.4 of 01 May 2000"; *)
     (*	VERSION    = "0.5 of 10 June 2000"; *)
     (*	VERSION    = "0.6 of 17 June 2000"; *)
     (*	VERSION    = "0.7 of 18 June 2000"; *)
     (*	VERSION    = "0.8 of 18 July 2000"; *)
     (*	VERSION    = "0.9 of 27 August 2000"; *)
     	VERSION    = "0.95 of 26 October 2000"; 
	verStr     = " version " + VERSION;
	prefix     = "#gpcp: ";
	millis     = "mSec";

(* ==================================================================== *)
(*		     State Variables of this compilation		*)
(* ==================================================================== *)

  VAR
    ntvObj* : Symbols.Type;	(* native Object type		*)
    ntvStr* : Symbols.Type;	(* native String type		*)
    ntvExc* : Symbols.Type;	(* native Exceptions type	*)

  VAR
    modNam*   : FileNames.NameString;	(* name of the _MODULE_		*)
    basNam-, 				(* base name of source _FILE_	*)
    srcNam-, 				(* name of the source file	*)
    lstNam-   : FileNames.NameString;	(* name of the listing file	*)
    target-   : ARRAY 4 OF CHAR;

    special-,
    warning-,
    verbose-,
    doStats-,
    doHelp-,
    ovfCheck-,
    doneHelp,
    doVersion-,
    doneVersion,
    doSym-,
    doAsm-,
    doJsmn-,
    doCode-,
    system-    : BOOLEAN;
    listLevel- : INTEGER;
    hashSize-  : INTEGER;

    thisMod-   : IdDesc.BlkId;		(* Desc. of compiling module.	*)

    impSeq*    : Symbols.ScpSeq;

    totalS*    : LONGINT;
    parseS*    : LONGINT;
    parseE*    : LONGINT;
    attrib*    : LONGINT;
    symEnd*    : LONGINT;
    asmEnd*    : LONGINT;
    totalE*    : LONGINT;
    import1*   : LONGINT;
    import2*   : LONGINT;

    impMax*    : INTEGER;


(* ==================================================================== *)
(*				Utilities				*)
(* ==================================================================== *)

    PROCEDURE Message*(IN mss : ARRAY OF CHAR);
    BEGIN
      Console.WriteString(prefix);
      Console.WriteString(mss);
      Console.WriteLn;
    END Message;

    PROCEDURE PrintLn*(IN mss : ARRAY OF CHAR);
    BEGIN
      Console.WriteString(mss);
      Console.WriteLn;
    END PrintLn;

    PROCEDURE ErrMesg*(IN mss : ARRAY OF CHAR);
    BEGIN
      Console.WriteString(prefix);
      Error.WriteString(mss);
      Error.WriteLn;
    END ErrMesg;

    PROCEDURE Abort*(IN mss : ARRAY OF CHAR);
    BEGIN
      ErrMesg(mss); ASSERT(FALSE);
    END Abort;

    PROCEDURE isForeign*() : BOOLEAN;
    BEGIN
      RETURN 
	(Symbols.rtsMd IN thisMod.xAttr) OR
	(Symbols.frnMd IN thisMod.xAttr);
    END isForeign;


    PROCEDURE TimeMsg(IN mss : ARRAY OF CHAR; tim : LONGINT);
    BEGIN
      IF (tim < 0) OR (tim >= totalS) THEN tim := 0 END;
      Console.WriteString(prefix);
      Console.WriteString(mss);
      Console.WriteInt(SHORT(tim), 5);
      Console.WriteString(millis);
      Console.WriteLn;
    END TimeMsg;

(* ==================================================================== *)

    PROCEDURE Usage;
    BEGIN
      Message("Usage from the command line ...");
      IF RTS.defaultTarget = "net" THEN
PrintLn("       $ gpcp [cp-options] file {file}");
PrintLn("# CP Options ...");
PrintLn("       /copyright   ==> Display copyright notice");
PrintLn("       /help        ==> Write out this usage message");
PrintLn("       /list        ==> (default) Create *.lst file if errors");
PrintLn("       /list+       ==> Unconditionally create *.lst file");
PrintLn("       /list-       ==> Don't create error *.lst file");
PrintLn("       /dostats     ==> Give a statistical summary");
PrintLn("       /nocode      ==> Don't create any object files");
PrintLn("       /noasm       ==> Don't create asm (or object) files");
PrintLn("       /nosym       ==> Don't create *.sym (or asm or object) files");
PrintLn("       /nocheck     ==> Don't perform arithmetic overflow checks");
PrintLn("       /verbose     ==> Emit verbose diagnostics");
PrintLn("       /version     ==> Write out version number");
PrintLn("       /warn-       ==> Don't emit warnings");
PrintLn("       /nowarn      ==> Don't emit warnings");
PrintLn("       /target=XXX  ==> Emit (jvm|net|dcf) assembly");
PrintLn("       /hsize=NNN   ==> Set hashtable size >= NNN (0 .. 65000)");
PrintLn("       /special     ==> Compile dummy symbol file");
PrintLn('	Unix-style options: "-option" are recognized also');
      ELSE
        IF RTS.defaultTarget = "jvm" THEN
PrintLn("       $ cprun gpcp [cp-options] file {file}, OR");
PrintLn("       $ java [java-options] CP.gpcp.gpcp [cp-options] file {file}");
        ELSIF RTS.defaultTarget = "dcf" THEN
	  PrintLn("       $ gpcp [cp-options] file {file}");
	END;
PrintLn("# CP Options ...");
PrintLn("       -copyright   ==> Display copyright notice");
PrintLn("       -help        ==> Write out this usage message");
PrintLn("       -list        ==> (default) Create *.lst file if errors");
PrintLn("       -list+       ==> Unconditionally create *.lst file");
PrintLn("       -list-       ==> Don't create error *.lst file");
PrintLn("       -dostats     ==> Give a statistical summary");
PrintLn("       -jasmin      ==> Ceate asm files and run Jasmin");
PrintLn("       -nocode      ==> Don't create any object files");
PrintLn("       -noasm       ==> Don't create asm (or object) files");
PrintLn("       -nosym       ==> Don't create *.sym (or asm or object) files");
PrintLn("       -verbose     ==> Emit verbose diagnostics");
PrintLn("       -version     ==> Write out version number");
PrintLn("       -warn-       ==> Don't emit warnings");
PrintLn("       -nowarn      ==> Don't emit warnings");
PrintLn("       -target=XXX  ==> Emit (jvm|net|dcf) assembly");
PrintLn("       -hsize=NNN   ==> Set hashtable size >= NNN (0 .. 65000)");
PrintLn("       -special     ==> Compile dummy symbol file");
        IF RTS.defaultTarget = "jvm" THEN
PrintLn("# Java Options ...");
PrintLn("       -D<name>=<value>  pass <value> to JRE as system property <name>");
PrintLn("       -DCPSYM=$CPSYM    pass value of CPSYM environment variable to JRE");
        END;
      END;
      Message("This program comes with NO WARRANTY");
      Message("Read source/COPYING for GNU GPL license details");
    END Usage;

(* ==================================================================== *)
(*			     Option Setting				*)
(* ==================================================================== *)

    PROCEDURE ParseOption*(IN opt : ARRAY OF CHAR);
      VAR copy : ARRAY 12 OF CHAR;
	  indx : INTEGER;
     (* ----------------------------------------- *)
      PROCEDURE Unknown(IN str : ARRAY OF CHAR);
      BEGIN
	Message('Unknown option "' + str + '"');
	doHelp := TRUE;
      END Unknown;
     (* ----------------------------------------- *)
      PROCEDURE BadSize();
      BEGIN Message('hsize must be integer in range 0 .. 65000') END BadSize;
     (* ----------------------------------------- *)
      PROCEDURE ParseSize(IN opt : ARRAY OF CHAR);
        VAR ix : INTEGER;
            nm : INTEGER;
            ch : CHAR;
      BEGIN
        nm := 0;
        ix := 7;
        WHILE opt[ix] # 0X DO
          ch := opt[ix];
          IF (ch >= '0') & (ch <= '9') THEN
            nm := nm * 10 + ORD(ch) - ORD('0');
            IF nm > 65521 THEN BadSize; hashSize := nm; RETURN END;
          ELSE
            BadSize; doHelp := TRUE; hashSize := nm; RETURN;
          END;
          INC(ix);
        END;
        hashSize := nm;
      END ParseSize;
     (* ----------------------------------------- *)
    BEGIN
      indx := 1;
      WHILE (indx < 12) & (indx < LEN(opt)) DO
	copy[indx-1] := opt[indx]; INC(indx);
      END;
      copy[11] := 0X;

      CASE copy[0] OF
      | "c" : IF copy = "copyright" THEN GPCPcopyright.Write END;
      | "d" : IF copy = "dostats" THEN doStats := TRUE ELSE Unknown(opt) END;
      | "w" : IF copy = "warn-" THEN warning := FALSE ELSE Unknown(opt) END;
      | "h" : 
          copy[6] := 0X;
          IF copy = "help" THEN
            doHelp := TRUE;
          ELSIF copy = "hsize=" THEN
            ParseSize(opt);
          ELSE
            Unknown(opt);
          END;
      | "j" :
	  IF copy = "jasmin" THEN
	    doCode     := TRUE;
	    doJsmn     := TRUE;
	  ELSE 
	    Unknown(opt);
	  END;
      | "v" :
	  IF copy = "version" THEN
	    doVersion  := TRUE;
	  ELSIF copy = "verbose" THEN
	    verbose    := TRUE;
	    doStats    := TRUE;
	    CPascalErrors.prompt := TRUE;
	  ELSE 
	    Unknown(opt);
	  END;
      | "n" :
	  IF copy = "nosym" THEN
	    doSym      := FALSE;
	    doAsm      := FALSE;
	    doCode     := FALSE;
	  ELSIF copy = "noasm" THEN
	    doAsm      := FALSE;
	    doCode     := FALSE;
	  ELSIF copy = "nocode" THEN
	    doCode     := FALSE;
	  ELSIF copy = "nowarn" THEN
	    warning    := FALSE;
	  ELSIF copy = "nocheck" THEN
	    ovfCheck   := FALSE;
	  ELSE 
	    Unknown(opt);
	  END;
      | "l" :
	  IF copy = "list-" THEN
	    listLevel  := CPascalS.listNever;
	  ELSIF copy = "list+" THEN
	    listLevel  := CPascalS.listAlways;
	  ELSIF copy = "list" THEN
	    listLevel  := CPascalS.listErrOnly;
	  ELSE 
	    Unknown(opt);
	  END;
      | "s" :
	  IF copy = "special" THEN
	    doAsm      := FALSE;
	    special    := TRUE;
  	  ELSE
	    Unknown(opt);
	  END;
      | "t" :
	  IF copy = "target=jvm" THEN
	    IF RTS.defaultTarget = "jvm" THEN
	      Message("JVM is default target for this build");
	    END;
	    target := "jvm";
	  ELSIF (copy = "target=vos") OR
		(copy = "target=net") THEN
	    IF RTS.defaultTarget = "net" THEN
	      Message("NET is default target for this build");
	    END;
	    target := "net";
	  ELSIF copy = "target=dcf" THEN
	    Message('DCode emitter not yet available, using "target=' +
		RTS.defaultTarget + '"');
	  ELSE 
	    Unknown(opt);
	  END;
      ELSE
	Unknown(opt);
      END;
      IF doVersion & ~doneVersion THEN 
	Message(target + verStr); 
	doneVersion := TRUE;
      END;
      IF doHelp & ~doneHelp THEN Usage; doneHelp := TRUE END;
    END ParseOption;

(* ==================================================================== *)

    PROCEDURE InitCompState*(IN nam : ARRAY OF CHAR);
    BEGIN
      IF verbose THEN Message("opened local file <" + nam + ">") END;
      GPText.Assign(nam, srcNam);
      FileNames.StripExt(nam, basNam);
      FileNames.AppendExt(basNam, "lst", lstNam);
      NEW(thisMod); 
      thisMod.SetKind(IdDesc.modId);
      thisMod.ovfChk := ovfCheck;
    END InitCompState;

(* ==================================================================== *)

  PROCEDURE Report*;
    VAR str1 : ARRAY 8 OF CHAR;
	str2 : ARRAY 8 OF CHAR;
  BEGIN
    Message(target + verStr); 
    GPText.IntToStr(CPascalS.line, str1);
    Message(str1 + " source lines");
    GPText.IntToStr(impMax, str1);
    Message("import recursion depth " + str1);
    GPText.IntToStr(NameHash.size, str2);
    GPText.IntToStr(NameHash.entries, str1);
    Message(str1 + " entries in hashtable of size " + str2);
    TimeMsg("import time   ", import2 - import1);
    TimeMsg("source time   ", parseS  - totalS);
    TimeMsg("parse time    ", parseE  - parseS - import2 + import1);
    TimeMsg("analysis time ", attrib  - parseE);
    TimeMsg("symWrite time ", symEnd  - attrib);
    TimeMsg("asmWrite time ", asmEnd  - symEnd);
    TimeMsg("assemble time ", totalE  - asmEnd);
    TimeMsg("total time    ", totalE  - totalS);
  END Report;

(* ==================================================================== *)

  PROCEDURE InitOptions*;
  BEGIN
    warning	:= TRUE;
    verbose	:= FALSE;
    doHelp	:= FALSE; doneHelp    := FALSE;
    doVersion	:= FALSE; doneVersion := FALSE;
    ovfCheck	:= TRUE;
    doSym	:= TRUE;
    doStats	:= FALSE;
    doJsmn 	:= FALSE;
    doCode	:= TRUE;
    doAsm 	:= TRUE;
    special 	:= FALSE;
    system 	:= FALSE;
    listLevel   := CPascalS.listErrOnly;
    hashSize    := 0;	(* gets default hash size *)
  END InitOptions;

(* ==================================================================== *)
BEGIN
  GPText.Assign(RTS.defaultTarget, target);
END CompState.
(* ==================================================================== *)

