(* ==================================================================== *)
(*									*)
(*  TypeDesc Module for the Gardens Point Component Pascal Compiler.	*)
(*  Implements type descriptors that are extensions of Symbols.Type	*)
(*									*)
(*	Copyright (c) John Gough 1999, 2000.				*)
(*									*)
(* ==================================================================== *)

MODULE TypeDesc;

  IMPORT 
	GPCPcopyright,
	Con := Console,
	GPText,
	VarSets,
	NameHash,
	FileNames,
	Id := IdDesc,
	G  := CompState,
	Sy := Symbols,
	Lv := LitValue,
	S  := CPascalS,
	H  := DiagHelper;

(* ============================================================ *)

  CONST (* type-kinds *)
    basTp* = Symbols.standard; 
    tmpTp* = 1; namTp* = 2; arrTp* = 3; 
    recTp* = 4; ptrTp* = 5; prcTp* = 6; 
    enuTp* = 7;

  CONST (* base-ordinals *)
    notBs   =  0;
    boolN*  =  1; 
    sChrN*  =  2; charN*  = 3; 
    byteN*  =  4; sIntN*  = 5; intN*  =  6; lIntN* =  7; 
    sReaN*  =  8; realN*  = 9; 
    setN*   = 10;
    anyRec* = 11; anyPtr* = 12; 
    strN*   = 13; sStrN*  = 14; metaN* = 15;

  CONST (* record attributes *)
    noAtt* = 0; isAbs* = 1; limit* = 2; extns* = 3; iFace* = 4;

(* ============================================================ *)

  CONST (* traversal depth markers *)
    initialMark = 0; 
    finishMark  = -1;
    errorMark   = 0FFFFFH;

(* ============================================================ *)

  TYPE
    Base*   = POINTER TO RECORD (Sy.Type)
	     (* ... inherited from Type ... ------------- *
	      *	idnt*   : Idnt;		(* Id of typename *)
	      *	kind-   : INTEGER;	(* tag for unions *)
	      *	serial- : INTEGER;	(* type serial-nm *)
	      * force*  : INTEGER;	(* force sym-emit *)
	      *	xName*  : Lv.CharOpen;  (* full ext name  *)
	      *	dump*,depth* : INTEGER; (* scratch loc'ns *)
	      * ----------------------------------------- *)
		tpOrd* : INTEGER;
	      END;	(* ------------------------------ *)

(* ============================================================ *)

  VAR anyRecTp- : Base;	(* Descriptor for base record type ANY. *)
      integerT* : Sy.Type;
      nilStr    : Lv.CharOpen;

(* ============================================================ *)

  TYPE
    Opaque* = POINTER TO RECORD (Sy.Type)
	     (* ... inherited from Type ... ------------- *
	      *	idnt*   : Idnt;		(* Id of typename *)
	      *	kind-   : INTEGER;	(* tag for unions *)
	      *	serial- : INTEGER;	(* type serial-nm *)
	      * force*  : INTEGER;	(* force sym-emit *)
	      *	xName*  : Lv.CharOpen;  (* full ext name  *)
	      *	dump*,depth* : INTEGER; (* scratch loc'ns *)
	      * ----------------------------------------- *)
		resolved* : Sy.Type;	(* ptr to real-Tp *)
	       	scopeNm*  : Lv.CharOpen;
	      END;	(* ------------------------------ *)

(* ============================================================ *)

  TYPE
    Array*  = POINTER TO RECORD (Sy.Type)
	     (* ... inherited from Type ... ------------- *
	      *	idnt*   : Idnt;		(* Id of typename *)
	      *	kind-   : INTEGER;	(* tag for unions *)
	      *	serial- : INTEGER;	(* type serial-nm *)
	      * force*  : INTEGER;	(* force sym-emit *)
	      *	xName*  : Lv.CharOpen;  (* full ext name  *)
	      *	dump*,depth* : INTEGER; (* scratch loc'ns *)
	      * ----------------------------------------- *)
		elemTp* : Sy.Type;	(* element tpDesc *)
		length* : INTEGER;	(* 0 for open-arr *)
	      END;	(* ------------------------------ *)

(* ============================================================ *)

  TYPE
    Record* = POINTER TO RECORD (Sy.Type)
	     (* ... inherited from Type ... ------------- *
	      *	idnt*   : Idnt;		(* Id of typename *)
	      *	kind-   : INTEGER;	(* tag for unions *)
	      *	serial- : INTEGER;	(* type serial-nm *)
	      * force*  : INTEGER;	(* force sym-emit *)
	      *	xName*  : Lv.CharOpen;  (* full ext name  *)
	      *	dump*,depth* : INTEGER; (* scratch loc'ns *)
	      * ----------------------------------------- *)
		baseTp*     : Sy.Type;	(* immediate base *)
		bindTp*     : Sy.Type;  (* ptrTo if anon. *)
		recAtt*     : INTEGER;
		symTb*	    : Sy.SymbolTable;
		extrnNm*    : Lv.CharOpen;
	       	scopeNm*    : Lv.CharOpen;
		fields*     : Sy.IdSeq;	(* list of fields *)
		methods*    : Sy.IdSeq;	(* list of meth's *)
		statics*    : Sy.IdSeq;	(* list of stat's *)
		interfaces* : Sy.IdSeq; (* impl-sequence  *)
		xAttr*      : SET;	(* external attrs *)
	      END;	(* ------------------------------ *)
(* ============================================================	*
 * When should a record type be implemented as a class?   	*
 * When any of the following is true:			  	*
 *	>  It is extensible or abstract			  	*
 *	>  It extends another record			  	*
 *	>  It has an embedded non-value structure	  	*
 *	>  It is only declared as a pointer target	  	*
 *	>  If the target does not support value records   	*
 * ============================================================	*)

(* ============================================================ *)

  TYPE
    Enum* =   POINTER TO RECORD (Sy.Type)
	     (* ... inherited from Type ... ------------- *
	      *	idnt*   : Idnt;		(* Id of typename *)
	      *	kind-   : INTEGER;	(* tag for unions *)
	      *	serial- : INTEGER;	(* type serial-nm *)
	      * force*  : INTEGER;	(* force sym-emit *)
	      *	xName*  : Lv.CharOpen;  (* full ext name  *)
	      *	dump*,depth* : INTEGER; (* scratch loc'ns *)
	      * ----------------------------------------- *)
		symTb*	    : Sy.SymbolTable;
		statics*    : Sy.IdSeq;	(* list of stat's *)
	      END;	(* ------------------------------ *)

(* ============================================================ *)

  TYPE
    Pointer*= POINTER TO RECORD (Sy.Type)
	     (* ... inherited from Type ... ------------- *
	      *	idnt*   : Idnt;		(* Id of typename *)
	      *	kind-   : INTEGER;	(* tag for unions *)
	      *	serial- : INTEGER;	(* type serial-nm *)
	      * force*  : INTEGER;	(* force sym-emit *)
	      *	xName*  : Lv.CharOpen;  (* full ext name  *)
	      *	dump*,depth* : INTEGER; (* scratch loc'ns *)
	      * ----------------------------------------- *)
		boundTp* : Sy.Type;	(* ptr bound type *)
	      END;	(* ------------------------------ *)

(* ============================================================ *)

  TYPE
    Procedure* = POINTER TO RECORD (Sy.Type)
	     (* ... inherited from Type ... ------------- *
	      *	idnt*   : Idnt;		(* Id of typename *)
	      *	kind-   : INTEGER;	(* tag for unions *)
	      *	serial- : INTEGER;	(* type serial-nm *)
	      * force*  : INTEGER;	(* force sym-emit *)
	      *	xName*  : Lv.CharOpen;  (* full ext name  *)
	      *	dump*,depth* : INTEGER; (* scratch loc'ns *)
	      * ----------------------------------------- *)
		retType*  : Sy.Type;	(* ret-type | NIL *)
		receiver* : Sy.Type;	(* element tpDesc *)
		formals*  : Id.ParSeq;	(* formal params  *)
		quasiPs*  : Id.ParSeq;	(* quasi params   *)
		retN*,argN* : INTEGER;
(*
		globLst : ProcVarGlob; ????
 *)
	      END;	(* ------------------------------ *)

(* ============================================================ *)
(*		   Predicates on Type extensions                *)
(* ============================================================ *)

  PROCEDURE (t : Base)isBooleanType*() : BOOLEAN;
  BEGIN RETURN t.tpOrd = boolN END isBooleanType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isNumType*() : BOOLEAN;
  BEGIN RETURN (t.tpOrd <= realN) & (t.tpOrd >= byteN) END isNumType;

  PROCEDURE (t : Enum)isNumType*() : BOOLEAN;
  BEGIN RETURN TRUE END isNumType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isBaseType*() : BOOLEAN;
  BEGIN RETURN TRUE END isBaseType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isIntType*() : BOOLEAN;
  BEGIN RETURN (t.tpOrd <= lIntN) & (t.tpOrd >= byteN) END isIntType;

  PROCEDURE (t : Enum)isIntType*() : BOOLEAN;
  BEGIN RETURN TRUE END isIntType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isScalarType*() : BOOLEAN; 
  BEGIN RETURN t.tpOrd # anyRec END isScalarType;

  PROCEDURE (t : Enum)isScalarType*() : BOOLEAN; 
  BEGIN RETURN TRUE END isScalarType;

  PROCEDURE (t : Array)isScalarType*() : BOOLEAN; 
  BEGIN RETURN FALSE END isScalarType;

  PROCEDURE (t : Record)isScalarType*() : BOOLEAN;
  BEGIN RETURN FALSE END isScalarType;

(* -------------------------------------------- *)

  PROCEDURE (t : Record)isImportedType*() : BOOLEAN;
  BEGIN
    IF t.bindTp # NIL THEN 
      RETURN t.bindTp.isImportedType();
    ELSE
      RETURN (t.idnt # NIL) & (t.idnt.dfScp # NIL) & t.idnt.dfScp.isImport();
    END;
  END isImportedType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isSetType*() : BOOLEAN;
  BEGIN RETURN t.tpOrd = setN END isSetType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isRealType*() : BOOLEAN;
  BEGIN RETURN (t.tpOrd = realN) OR (t.tpOrd = sReaN) END isRealType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isCharType*() : BOOLEAN;
  BEGIN RETURN (t.tpOrd = charN) OR (t.tpOrd = sChrN) END isCharType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isStringType*() : BOOLEAN;
  BEGIN RETURN (t.tpOrd = strN) OR (t.tpOrd = sStrN) END isStringType;

(* -------------------------------------------- *)

  PROCEDURE (t : Pointer)nativeCompat*() : BOOLEAN;
  BEGIN RETURN (t = G.ntvStr) OR (t = G.ntvObj) END nativeCompat;

(* -------------------------------------------- *)

  PROCEDURE (t : Array)isCharArrayType*() : BOOLEAN;
  BEGIN RETURN (t.elemTp # NIL) & t.elemTp.isCharType() END isCharArrayType;

(* -------------------------------------------- *)

  PROCEDURE (t : Pointer)isDynamicType*() : BOOLEAN;
  (** A type is dynamic if it is a pointer to any 		*
   *  record. Overrides isDynamicType method in Symbols.Type.	*)
  BEGIN RETURN (t.boundTp # NIL) & (t.boundTp IS Record) END isDynamicType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isPointerType*() : BOOLEAN;
  BEGIN RETURN t.tpOrd = anyPtr END isPointerType;

(* -------------------------------------------- *)

  PROCEDURE (t : Pointer)isPointerType*() : BOOLEAN;
  BEGIN RETURN TRUE END isPointerType;

(* -------------------------------------------- *)

  PROCEDURE (t : Procedure)isProcType*() : BOOLEAN;
  BEGIN RETURN TRUE END isProcType;

(* -------------------------------------------- *)

  PROCEDURE (t : Procedure)isProperProcType*() : BOOLEAN;
  BEGIN RETURN t.retType = NIL END isProperProcType;

(* -------------------------------------------- *)

  PROCEDURE (t : Procedure)returnType*() : Sy.Type;
  BEGIN RETURN t.retType END returnType;

(* -------------------------------------------- *)

  PROCEDURE (t : Record)isAbsRecType*() : BOOLEAN;
  (** A record is absolute if it is declared to be an absolute	*
   *  record. Overrides isAbsRecType method in Symbols.Type.	*)
  BEGIN 
    RETURN (isAbs = t.recAtt) OR (iFace = t.recAtt);
  END isAbsRecType;

(* -------------------------------------------- *)

  PROCEDURE (t : Record)isExtnRecType*() : BOOLEAN;
  (** A record is extensibel if declared absolute or extensible	*
   *  record. Overrides isAbsRec method in Symbols.Type.	*)
  BEGIN 
    RETURN (extns = t.recAtt) OR (isAbs = t.recAtt);
  END isExtnRecType;

(* -------------------------------------------- *)

  PROCEDURE (t : Record)isLimRecType*() : BOOLEAN;
  (** A record is limited if it is declared to be limited	*
   *  record. Overrides isLimRec method in Symbols.Type.	*)
  BEGIN RETURN limit = t.recAtt END isLimRecType;

(* -------------------------------------------- *)

  PROCEDURE (t : Array)isOpenArrType*() : BOOLEAN;
  BEGIN RETURN t.length = 0 END isOpenArrType;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)needsInit*() : BOOLEAN;
  BEGIN RETURN FALSE END needsInit;

(* -------------------------------------------- *)

  PROCEDURE (t : Base)isLongType*() : BOOLEAN;
  BEGIN RETURN (t.tpOrd = realN) OR (t.tpOrd = lIntN) END isLongType;

(* -------------------------------------------- *)

  PROCEDURE (b : Base)includes*(x : Sy.Type) : BOOLEAN;
    VAR xBas : Base;
  BEGIN 
    IF x IS Enum THEN x := integerT;
    ELSIF ~(x IS Base) THEN RETURN FALSE;
    END;

    xBas := x(Base);
    CASE b.tpOrd OF
    | byteN .. realN : 
	RETURN (xBas.tpOrd <= b.tpOrd) & (xBas.tpOrd >= byteN);
    | charN :
	RETURN (xBas.tpOrd = charN) OR (xBas.tpOrd = sChrN);
    | sChrN :
	RETURN xBas.tpOrd = sChrN;
    ELSE
	RETURN FALSE;
    END;
  END includes;

  PROCEDURE (b : Enum)includes*(x : Sy.Type) : BOOLEAN;
    VAR xBas : Base;
  BEGIN 
    RETURN integerT.includes(x);
  END includes;

(* -------------------------------------------- *)

  PROCEDURE (b : Base)isBaseOf*(e : Sy.Type) : BOOLEAN;
  (** Find if e is an extension of base type b.		*
   *  Overrides the isBaseOf method in Symbols.Type		*)
  BEGIN
    IF    e.kind = recTp THEN RETURN b.tpOrd = anyRec;
    ELSIF e.kind = ptrTp THEN RETURN b.tpOrd = anyPtr;
    ELSE (* all others *)     RETURN b = e;
    END;
  END isBaseOf;

(* -------------------------------------------- *)

  PROCEDURE (b : Record)isBaseOf*(e : Sy.Type) : BOOLEAN;
  (** Find if e is an extension of record type b.		*
   *  Overrides the isBaseOf method in Symbols.Type		*)
    VAR ext : Record;
  BEGIN
    IF (e = NIL) OR (e.kind # recTp) THEN RETURN FALSE;
    ELSIF e = b THEN RETURN TRUE;		(* Trivially!	*)
    END;					(* Not a record *)
    ext := e(Record);				(* Cast to Rec. *)
    RETURN b.isBaseOf(ext.baseTp);		(* Recurse up!  *)
  END isBaseOf;

(* -------------------------------------------- *)

  PROCEDURE (b : Pointer)isBaseOf*(e : Sy.Type) : BOOLEAN;
  (** Find if e is an extension of pointer type b.		*
   *  Overrides the isBaseOf method in Symbols.Type		*)
    VAR ext : Pointer;
  BEGIN
    IF (e = NIL) OR (e.kind # ptrTp) THEN RETURN FALSE;
    ELSIF e = b THEN RETURN TRUE;		(* Trivially!	*)
    END;
    ext := e(Pointer);				(* Cast to Ptr. *)
    RETURN (b.boundTp # NIL)			(* Go to bnd-tp *)
	  & b.boundTp.isBaseOf(ext.boundTp);	(* for decision	*)
  END isBaseOf;

(* -------------------------------------------- *)

  PROCEDURE (lhT : Array)arrayCompat*(rhT : Sy.Type) : BOOLEAN;
  BEGIN 
    IF lhT.length = 0 THEN			(* An open array type	*)
      IF rhT.kind = arrTp THEN
	RETURN lhT.elemTp.arrayCompat(rhT(Array).elemTp);
      ELSE
	RETURN lhT.isCharArrayType() & rhT.isStringType();
      END;
    ELSE
      RETURN FALSE;
    END;
  END arrayCompat;

(* -------------------------------------------- *)

  PROCEDURE (lhT : Enum)equalType*(rhT : Sy.Type) : BOOLEAN;
  BEGIN
    IF lhT = rhT THEN RETURN TRUE END;
    WITH rhT : Base DO 
      RETURN rhT = integerT;
    ELSE 
      RETURN FALSE;
    END;
  END equalType;

(* -------------------------------------------- *)
(*
 *PROCEDURE (t : Sy.Type)isExtJava*() : BOOLEAN;
 *BEGIN RETURN FALSE END isExtJava;
 *)
(* -------------------------------------------- *)
(* Used for unwriteable extensions of Java classes
  PROCEDURE (t : XXXX)valCopyOK*() : BOOLEAN;
  BEGIN RETURN TRUE 
  END valCopyOK;
 *)
(* -------------------------------------------- *)

  PROCEDURE (t : Record)isInterfaceType*() : BOOLEAN;
    CONST mask = {Sy.jvInf, Sy.csInf};
  BEGIN 
    RETURN t.recAtt = iFace;
  END isInterfaceType;

(* -------------------------------------------- *)

  PROCEDURE (t : Pointer)isInterfaceType*() : BOOLEAN;
  BEGIN 
    RETURN (t.boundTp # NIL) & t.boundTp.isInterfaceType();
  END isInterfaceType;

(* -------------------------------------------- *)

  PROCEDURE (lhsT : Procedure)formsMatch(rhsT : Procedure) : BOOLEAN,NEW;
    VAR index : INTEGER;
	lP,rP : Id.ParId;
  BEGIN
    IF lhsT.formals.tide # rhsT.formals.tide THEN RETURN FALSE;
    ELSE
      FOR index := 0 TO lhsT.formals.tide-1 DO
	lP := lhsT.formals.a[index];
	rP := rhsT.formals.a[index];
	IF (lP.type # NIL) & ~lP.type.equalType(rP.type) THEN RETURN FALSE END;
	IF lP.parMod # rP.parMod THEN RETURN FALSE END;
      END;
    END;
    RETURN TRUE;
  END formsMatch;

(* -------------------------------------------- *)

  PROCEDURE (lT : Array)equalOpen*(r : Sy.Type) : BOOLEAN;
    VAR rT : Array;
  BEGIN
    IF ~(r IS Array) THEN RETURN FALSE;
    ELSE
      rT := r(Array);
      RETURN (lT.length = 0) & 
	     (rT.length = 0) &
	     lT.elemTp.equalType(rT.elemTp);
    END;
  END equalOpen;

(* -------------------------------------------- *)

  PROCEDURE (lT : Pointer)equalPointers*(r : Sy.Type) : BOOLEAN;
    VAR rT : Pointer;
  BEGIN
    IF ~(r IS Pointer) THEN RETURN FALSE;
    ELSE
      rT := r(Pointer);
      RETURN lT.boundTp.equalType(rT.boundTp);
    END;
  END equalPointers;

(* -------------------------------------------- *)

  PROCEDURE (i : Record)InstantiateCheck*(tok : S.Token),NEW;
  BEGIN
    IF i.recAtt = isAbs THEN
      S.SemError.Report(90, tok.lin, tok.col);
    ELSIF (i.idnt # NIL) & i.idnt.isImported() & 
	((i.recAtt = limit) OR (i.recAtt = iFace)) THEN
      S.SemError.Report(71, tok.lin, tok.col);
    (* check noArgOk for JVM version? ... error 155 *)
    END;
  END InstantiateCheck;

(* -------------------------------------------- *)

  PROCEDURE (lhsT : Procedure)procMatch*(rT : Sy.Type) : BOOLEAN;
    VAR rhsT : Procedure;
  BEGIN
    IF ~(rT IS Procedure) THEN RETURN FALSE;
    ELSE
      rhsT := rT(Procedure);
      IF (lhsT.retType = NIL) # (rhsT.retType = NIL) THEN RETURN FALSE END;
      IF (lhsT.retType # NIL) & 
	 ~lhsT.retType.equalType(rhsT.retType) THEN RETURN FALSE END;
      RETURN lhsT.formsMatch(rhsT);
    END;
  END procMatch;

(* -------------------------------------------- *)

  PROCEDURE (lhsT : Procedure)namesMatch*(rT : Sy.Type) : BOOLEAN;
    VAR rhsT  : Procedure;
	index : INTEGER;
  BEGIN
    IF ~(rT IS Procedure) THEN  RETURN FALSE;
    ELSE
      rhsT := rT(Procedure);
      IF lhsT.formals.tide # rhsT.formals.tide THEN RETURN FALSE END;
      FOR index := 0 TO lhsT.formals.tide-1 DO
	IF lhsT.formals.a[index].hash # 
	   rhsT.formals.a[index].hash THEN RETURN FALSE END;
      END;
      RETURN TRUE;
    END;
  END namesMatch;

(* -------------------------------------------- *)

  PROCEDURE (oldT : Procedure)CheckCovariance*(newI : Sy.Idnt);
  (* When a method is overidden, the formals must match, except *
   * that the return type may vary covariantly with the recvr.	*)
    VAR newT : Procedure;
  BEGIN
    IF newI IS Id.Procs THEN
      newT := newI.type(Procedure);
      IF (oldT.retType = NIL) # (newT.retType = NIL) THEN newI.IdError(116);
      ELSIF ~newT.formsMatch(oldT) THEN
	newI.IdError(160);
(*
 *    ELSIF (oldT.retType # NIL) &
 *	    (oldT.retType # newT.retType) &
 *	    ~oldT.retType.isBaseOf(newT.retType) THEN
 *	Sy.RepTypesError(116, oldT, newT, newI.token.lin, newI.token.col);
 *)
      ELSIF (oldT.retType # NIL) & (oldT.retType # newT.retType) THEN
	IF ~oldT.retType.isBaseOf(newT.retType) THEN
	  Sy.RepTypesError(116, oldT, newT, newI.token.lin, newI.token.col);
	ELSIF newI IS Id.MthId THEN
	  INCL(newI(Id.MthId).mthAtt, Id.covar);
	END;
      END; 
    END;
  END CheckCovariance;

(* -------------------------------------------- *)

  PROCEDURE (desc : Procedure)CheckEmptyOK*();
    VAR idx : INTEGER;
        frm : Id.ParId;
  BEGIN
    FOR idx := 0 TO desc.formals.tide - 1 DO
      frm := desc.formals.a[idx];
      IF frm.parMod = Sy.out THEN frm.IdError(114) END;
    END;
    IF desc.retType # NIL THEN desc.TypeError(115) END;
  END CheckEmptyOK;

(* -------------------------------------------- *)

  PROCEDURE (recT : Record)bindField*(hash : INTEGER) : Sy.Idnt,NEW;
    VAR idnt : Symbols.Idnt;
  BEGIN
    idnt := recT.symTb.lookup(hash);
    WHILE (idnt = NIL) & 			(* while not found yet	*)
	  (recT.baseTp # NIL) & 		(* while known base	*)
	  (recT.baseTp # anyRecTp) DO		(* while base not ANY	*)
      recT := recT.baseTp(Record);
      idnt := recT.symTb.lookup(hash);
    END;
    RETURN idnt;     
  END bindField;

(* -------------------------------------------- *)

  PROCEDURE (desc : Procedure)OutCheck*(v : VarSets.VarSet; eNo : INTEGER);
    VAR idx : INTEGER;
        frm : Id.ParId;
	msg : POINTER TO FileNames.NameString;
  BEGIN
    msg := NIL;
    FOR idx := 0 TO desc.formals.tide - 1 DO
      frm := desc.formals.a[idx];
      IF (frm.parMod = Sy.out) & ~v.includes(frm.varOrd) THEN 
	IF msg = NIL THEN
	  NEW(msg);
	  Sy.getName.Of(frm, msg);
	ELSE
	  GPText.Assign(msg^ + "," + Sy.getName.ChPtr(frm)^, msg);
	END;
      END;
    END;
    IF msg # NIL THEN desc.TypeErrStr(eNo, msg) END;
  END OutCheck;

(* ============================================================ *)
(* 		   Record error reporting methods		*)
(* ============================================================ *)

  PROCEDURE (ty : Record)TypeError*(n : INTEGER);
  BEGIN
    IF ty.bindTp # NIL THEN 
      ty.bindTp.TypeError(n);
    ELSE
      ty.TypeError^(n);
    END;
  END TypeError;

(* -------------------------------------------- *)

  PROCEDURE (ty : Record)TypeErrStr*(n : INTEGER; 
				  IN s : ARRAY OF CHAR);
  BEGIN
    IF ty.bindTp # NIL THEN 
      ty.bindTp.TypeErrStr(n,s);
    ELSE
      ty.TypeErrStr^(n,s);
    END;
  END TypeErrStr;

(* ============================================================ *)
(* 			Constructor methods			*)
(* ============================================================ *)

  PROCEDURE newBasTp*() : Base;
    VAR rslt : Base;
  BEGIN
    NEW(rslt);
    rslt.SetKind(basTp);
    RETURN rslt;
  END newBasTp;

(* ---------------------------- *)

  PROCEDURE newNamTp*() : Opaque;
    VAR rslt : Opaque;
  BEGIN
    NEW(rslt);
    rslt.SetKind(namTp);
    RETURN rslt;
  END newNamTp;

(* ---------------------------- *)

  PROCEDURE newTmpTp*() : Opaque;
    VAR rslt : Opaque;
  BEGIN
    NEW(rslt);
    rslt.SetKind(tmpTp);
    RETURN rslt;
  END newTmpTp;

(* ---------------------------- *)

  PROCEDURE newArrTp*() : Array;
    VAR rslt : Array;
  BEGIN
    NEW(rslt);
    rslt.SetKind(arrTp);
    RETURN rslt;
  END newArrTp;

  PROCEDURE mkArrayOf*(e : Symbols.Type) : Array;
    VAR rslt : Array;
  BEGIN
    NEW(rslt);
    rslt.SetKind(arrTp);
    rslt.elemTp := e;
    RETURN rslt;
  END mkArrayOf;

(* ---------------------------- *)

  PROCEDURE newRecTp*() : Record;
    VAR rslt : Record;
  BEGIN
    NEW(rslt);
    rslt.SetKind(recTp);
    RETURN rslt;
  END newRecTp;

(* ---------------------------- *)

  PROCEDURE newEnuTp*() : Enum;
    VAR rslt : Enum;
  BEGIN
    NEW(rslt);
    rslt.SetKind(enuTp);
    RETURN rslt;
  END newEnuTp;

(* ---------------------------- *)

  PROCEDURE newPtrTp*() : Pointer;
    VAR rslt : Pointer;
  BEGIN
    NEW(rslt);
    rslt.SetKind(ptrTp);
    RETURN rslt;
  END newPtrTp;

(* ---------------------------- *)

  PROCEDURE newPrcTp*() : Procedure;
    VAR rslt : Procedure;
  BEGIN
    NEW(rslt);
    rslt.SetKind(prcTp);
    RETURN rslt;
  END newPrcTp;

(* ============================================================ *)
(*		      Some Helper procedures			*)
(* ============================================================ *)

  PROCEDURE coverType*(a,b : Sy.Type) : Sy.Type;
  BEGIN
    IF    a.includes(b) THEN RETURN a;
    ELSIF b.includes(a) THEN RETURN b;
    ELSE RETURN NIL;
    END;
  END coverType;

(* ---------------------------- *)

  PROCEDURE Error145(start : Sy.Type);
    VAR sccTab : Sy.SccTable;
  BEGIN
    NEW(sccTab);
    sccTab.target := start;
    start.SccTab(sccTab);
    start.TypeErrStr(145, Sy.dumpList(sccTab.symTab));
  END Error145;

(* ============================================================ *)
(* 		Implementation of Abstract methods		*)
(* ============================================================ *)

  PROCEDURE (i : Base)resolve*(d : INTEGER) : Sy.Type;
  BEGIN RETURN i END resolve;

(* ---------------------------- *)

  PROCEDURE (i : Enum)resolve*(d : INTEGER) : Sy.Type;
  BEGIN RETURN i END resolve;

(* ---------------------------- *)

  PROCEDURE (i : Opaque)resolve*(d : INTEGER) : Sy.Type;
    VAR newTpId : Sy.Idnt;
	oldTpId : Sy.Idnt;
  BEGIN 
    IF i.depth = initialMark THEN 
     (*
      *  If i.kind=tmpTp, this is a forward type, or 
      *  a sym-file temporary.  If we cannot resolve
      *  this to a real type, it is an error.
      *
      *  If i.kind=namTp, this is a named opaque type, 
      *  we must look it up in the symTab.  If we
      *  do not find it, the type just stays opaque.
      *)
      i.depth := finishMark;
      oldTpId := i.idnt;
      newTpId := oldTpId.dfScp.symTb.lookup(oldTpId.hash);
      IF newTpId = NIL THEN 
	oldTpId.IdError(2);
      ELSIF newTpId.kind # Id.typId THEN 
	oldTpId.IdError(5);
      ELSIF newTpId.type # NIL THEN
       (*
	*   This particular method might recurse, even for
	*   correct programs, such as
	*	TYPE A = POINTER TO B;
	*	TYPE B = RECORD c : C END;
	*	TYPE C = RECORD(A) ... END;
	*   Thus we must not recurse until we have set the
	*   resolved field, since we have now set the depth
	*   mark and will not reenter the binding code again.
	*)
	i.resolved := newTpId.type;
	i.resolved := newTpId.type.resolve(d);	(* Recurse! *)
	IF i.kind = tmpTp THEN
	  IF i.resolved = i THEN oldTpId.IdError(125) END;
	ELSIF i.kind = namTp THEN
	  IF (i.resolved = NIL) OR
	     (i.resolved.kind = namTp) THEN i.resolved := i END;
        END;
      END;
    END;
    RETURN i.resolved;
  END resolve;

(* ---------------------------- *)

  PROCEDURE (i : Array)resolve*(d : INTEGER) : Sy.Type;
    VAR e137,e145 : BOOLEAN;
  BEGIN 
    IF i.depth = initialMark THEN 
      e145 := FALSE;
      e137 := FALSE;
      i.depth := d;
      IF i.elemTp # NIL THEN i.elemTp := i.elemTp.resolve(d) END;
      IF (i.length # 0) & 
	 (i.elemTp # NIL) & 
	  i.elemTp.isOpenArrType() THEN
	i.TypeError(69);
      END;
      IF i.depth = errorMark THEN 
	IF i.elemTp = i THEN e137 := TRUE ELSE e145 := TRUE END;
	i.TypeError(126);
      END;
      i.depth := finishMark;
      IF    e145 THEN Error145(i);
      ELSIF e137 THEN i.TypeError(137);
      END;
    ELSIF i.depth = d THEN (* recursion through value types *)
      i.depth := errorMark;
    END;
    RETURN i;
  END resolve;

(* ---------------------------- *)

  PROCEDURE (x: Record)CopyFieldsOf(b : Sy.Type),NEW; (* final *) 
    VAR bRecT : Record;
	nextF : Sy.Idnt; 
	index : INTEGER;
  BEGIN
    IF (b # anyRecTp) & (b.depth # errorMark) THEN
      bRecT := b.boundRecTp()(Record);
     (* 
      *   First get the fields of the higher ancestors. 
      *)
      IF bRecT.baseTp # NIL THEN x.CopyFieldsOf(bRecT.baseTp) END;
     (*
      *   Now add the fields of the immediate base type 
      *)
      FOR index := 0 TO bRecT.fields.tide-1 DO
	nextF := bRecT.fields.a[index];
        IF ~x.symTb.enter(nextF.hash, nextF) THEN
	  x.symTb.lookup(nextF.hash).IdError(82);
	END;
      END;
    END;
  END CopyFieldsOf;

(* ---------------------------- *)

  PROCEDURE (i : Record)resolve*(d : INTEGER) : Sy.Type;
  (** Resolve this type, and any used in this type *)
    VAR origT : Sy.Type;
	field : Sy.Idnt;
	index : INTEGER;
	hashN : INTEGER;
	nameS : Lv.CharOpen;
	ident : Sy.Idnt;
	recId : Sy.Idnt;
	e137,e145 : BOOLEAN;
   (* -----------------------------------------	*)
    PROCEDURE refInNET(t : Sy.Type) : BOOLEAN;
     (*
      *  This predicate is used for the .NET 
      *  platform, to set the "clsTp" attribute.
      *  It implies that this type will have a
      *  reference representation in .NET
      *)
    BEGIN
      IF t = NIL THEN 
		RETURN FALSE;	(* Actually we don't care here. *)
      ELSE WITH t : Record DO
		RETURN Sy.clsTp IN t.xAttr;
	   | t : Array DO
	        RETURN TRUE;	(* arrays are references in NET *)
	   ELSE RETURN FALSE;	(* all others are value types.	*)
      END  END;
    END refInNET;
   (* -----------------------------------------	*)
  BEGIN
    IF i.depth = initialMark THEN 
      IF i.idnt # NIL THEN
	recId := i.idnt;
        nameS := Sy.getName.ChPtr(recId);
      END;

      i.depth := d;
      e145 := FALSE;
      e137 := FALSE;
     (*
      *  First: resolve the base type, if any,
      *  or set the base type to the type ANY.
      *)
      IF i.baseTp = NIL THEN 
	i.baseTp := anyRecTp; 
      ELSIF i.baseTp # anyRecTp THEN
	origT := i.baseTp;
	i.baseTp := i.baseTp.resolve(d);
	IF i.baseTp # NIL THEN
	  i.baseTp := i.baseTp.boundRecTp();
	  IF i.baseTp = NIL THEN 
	    i.TypeError(14);	(* not rec or ptr *)
	  END;
	  IF i.depth = errorMark THEN 
	    IF i.baseTp = i THEN e137 := TRUE ELSE e145 := TRUE END;
	    i.TypeError(123); 
	  END;
        END;
	IF i.baseTp # NIL THEN 
	 (*
	  *  Base is resolved, now check some semantic constraints.
	  *)
	  IF (isAbs = i.recAtt) & ~i.baseTp.isAbsRecType() THEN
	    i.TypeError(102);	(* abstract record must have abstract base *)
	  ELSIF i.baseTp.isExtnRecType() THEN
	    i.CopyFieldsOf(i.baseTp);
	  ELSE
	    i.TypeError(16);	(* base type is not an extensible record   *)
	  END;
	END;
      END;
     (*
      *   Now check semantics of interface implementation.
      *)
      IF i.interfaces.tide > 0 THEN 
	FOR index := 0 TO i.interfaces.tide-1 DO
	  field := i.interfaces.a[index];
	  ASSERT(field.kind = Id.typId);
	  IF field.type # NIL THEN
	    field.type := field.type.boundRecTp();
	    IF (field.type # NIL) & 
	        ~field.type.isInterfaceType() THEN
	      i.TypeErrStr(158, field.type.name());
	    END;
	  END;
	END;
      END;
      i.depth := d;
     (*
      *  Next: resolve all field types.
      *)
      FOR index := 0 TO i.fields.tide-1 DO
	field := i.fields.a[index];
        IF field.type # NIL THEN field.type := field.type.resolve(d) END;
	IF i.depth = errorMark THEN 
	  IF field.type = i THEN e137 := TRUE ELSE e145 := TRUE END;
	  field.IdError(124);
	  i.depth := d;
	END;
	IF refInNET(field.type) THEN INCL(i.xAttr,Sy.clsTp) END;
      END;
     (*
      *  Next: resolve types of all static members.
      *)
      FOR index := 0 TO i.statics.tide-1 DO
	field := i.statics.a[index];
        IF field.type # NIL THEN field.type := field.type.resolve(d) END;
	IF i.depth = errorMark THEN 
	  IF field.type = i THEN e137 := TRUE ELSE e145 := TRUE END;
	  field.IdError(124);
	  i.depth := d;
	END;
      END;

      i.depth := finishMark;
      IF    e145 THEN Error145(i);
      ELSIF e137 THEN i.TypeError(137);
      END;
    ELSIF i.depth = d THEN (* recursion through value types *)
      i.depth := errorMark;
    END;
    RETURN i;
  END resolve;

(* ---------------------------- *)

  PROCEDURE (i : Pointer)resolve*(d : INTEGER) : Sy.Type;
  BEGIN
    IF i.depth = initialMark THEN 
      i.depth := d;
      IF (i.boundTp # NIL) & 		(*==> bound type is OK	*)
	 (i.boundTp IS Record) &	(*==> boundTp is record *)
	 (i.boundTp.idnt = NIL) THEN 	(*==> anon. bound type	*)
	INCL(i.boundTp(Record).xAttr, Sy.clsTp);
	IF i.idnt # NIL THEN		(*==> named ptr type	*)
         (*
	  *  The anon record should have the same name as the
	  *  pointer type.  The record is marked so that the
	  *  synthetic name <ptrName>"^" can be derived.
	  *  The visibility mode is the same as the pointer.
	  *)
	  i.boundTp.idnt := NIL;		(* Zap synthetic name	*)
	  INCL(i.boundTp(Record).xAttr, Sy.anon);
	  i.boundTp(Record).bindTp := i;
	END;
      END;
      IF i.boundTp # NIL THEN i.boundTp := i.boundTp.resolve(d+1) END;
      i.depth := finishMark;
    END;
    RETURN i;
  END resolve;

(* ---------------------------- *)

  PROCEDURE (i : Procedure)resolve*(d : INTEGER) : Sy.Type;
    VAR idx : INTEGER;
        frm : Sy.Idnt;
  BEGIN
    IF i.depth = initialMark THEN 
      i.depth := d;
      FOR idx := 0 TO i.formals.tide-1 DO
        frm := i.formals.a[idx];
        IF frm.type # NIL THEN frm.type := frm.type.resolve(d+1) END;
      END;

      IF i.retType # NIL THEN i.retType := i.retType.resolve(d+1) END;
      i.depth := finishMark;
    END;
    RETURN i
  END resolve;

(* ============================================================ *)

  PROCEDURE Insert(VAR s : Sy.SymbolTable; t : Sy.Type);
    VAR junk : BOOLEAN;
  BEGIN
    IF t.idnt # NIL THEN junk := s.enter(t.idnt.hash, t.idnt) END;
  END Insert;

(* ----------------------------------------------------	*)

  PROCEDURE (i : Array)SccTab*(t : Sy.SccTable);
  BEGIN
    t.reached := FALSE;
    IF i.elemTp # NIL THEN
      IF i.elemTp = t.target THEN
	t.reached := TRUE;
      ELSE
        i.elemTp.SccTab(t);
      END;
      IF t.reached THEN Insert(t.symTab, i) END;
    END;
  END SccTab;

(* ----------------------------------------------------	*)

  PROCEDURE (i : Record)SccTab*(t : Sy.SccTable);
    VAR index : INTEGER;
	found : BOOLEAN;
	field : Sy.Idnt;
  BEGIN
    found := FALSE;
    IF i.baseTp # NIL THEN
      IF i.baseTp = t.target THEN
        t.reached := TRUE;
      ELSE
        t.reached := FALSE;
        i.baseTp.SccTab(t);
      END;
      IF t.reached THEN found := TRUE END;
    END;
    FOR index := 0 TO i.fields.tide-1 DO
      field := i.fields.a[index];
      IF field.type # NIL THEN
        IF field.type = t.target THEN
          t.reached := TRUE;
        ELSIF field.type # NIL THEN
          t.reached := FALSE;
          field.type.SccTab(t);
        END;
        IF t.reached THEN found := TRUE END;
      END;
    END;
    IF found THEN Insert(t.symTab, i); t.reached := TRUE END;
  END SccTab;

(* ----------------------------------------------------	*)

  PROCEDURE (i : Base)SccTab*(t : Sy.SccTable); 
  BEGIN (* skip *) END SccTab;

(* ----------------------------------------------------	*)

  PROCEDURE (i : Opaque)SccTab*(t : Sy.SccTable); 
  BEGIN (* skip *) END SccTab;

(* ----------------------------------------------------	*)

  PROCEDURE (i : Pointer)SccTab*(t : Sy.SccTable); 
  BEGIN (* skip *) END SccTab;

(* ----------------------------------------------------	*)

  PROCEDURE (i : Enum)SccTab*(t : Sy.SccTable); 
  BEGIN (* skip *) END SccTab;

(* ----------------------------------------------------	*)

  PROCEDURE (i : Procedure)SccTab*(t : Sy.SccTable);
  BEGIN (* skip *) END SccTab;

(* ============================================================ *)

  PROCEDURE update*(IN a : Sy.TypeSeq; t : Sy.Type) : Sy.Type;
  BEGIN
    IF t.kind = tmpTp THEN RETURN a.a[t.dump - Sy.tOffset] ELSE RETURN t END; 
  END update;

(* ============================================================ *)

  PROCEDURE (t : Base)TypeFix*(IN a : Sy.TypeSeq);
  BEGIN END TypeFix;

(* ---------------------------- *)

  PROCEDURE (t : Enum)TypeFix*(IN a : Sy.TypeSeq);
  BEGIN END TypeFix;

(* ---------------------------- *)

  PROCEDURE (t : Opaque)TypeFix*(IN a : Sy.TypeSeq);
  BEGIN END TypeFix;

(* ---------------------------- *)

  PROCEDURE (t : Array)TypeFix*(IN a : Sy.TypeSeq);
  BEGIN 
    t.elemTp := update(a, t.elemTp);
  END TypeFix;

(* ---------------------------- *)

  PROCEDURE (t : Record)TypeFix*(IN a : Sy.TypeSeq);
    VAR i : INTEGER;
	f : Sy.Idnt;
	m : Id.MthId;
  BEGIN 
    IF t.baseTp # NIL THEN t.baseTp := update(a, t.baseTp) END;
    FOR i := 0 TO t.fields.tide - 1 DO
      f := t.fields.a[i];
      f.type := update(a, f.type);
    END;
    FOR i := 0 TO t.methods.tide - 1 DO
      f := t.methods.a[i];
      m := f(Id.MthId);
      m.bndType := update(a, m.bndType);
      m.rcvFrm.type := update(a, m.rcvFrm.type);
      f.type.TypeFix(a);	(* recurse to param-types etc.	*)
    END;
    FOR i := 0 TO t.statics.tide - 1 DO
      f := t.statics.a[i];
      f.type := update(a, f.type);
      f.type.TypeFix(a);	(* recurse to param-types etc.	*)
    END;
  END TypeFix;

(* ---------------------------- *)

  PROCEDURE (t : Pointer)TypeFix*(IN a : Sy.TypeSeq);
    VAR bndT : Sy.Type;
  BEGIN 
    bndT := update(a, t.boundTp);
    t.boundTp := bndT;
    IF  bndT.idnt = NIL THEN
      WITH bndT : Record DO
	bndT.bindTp := t;
	INCL(bndT.xAttr, Sy.clsTp);
      ELSE (* ptr to array : skip *)
      END; 
    END;
  END TypeFix;

(* ---------------------------- *)

  PROCEDURE (t : Procedure)TypeFix*(IN a : Sy.TypeSeq);
    VAR i : INTEGER;
	f : Id.ParId;
  BEGIN 
    IF t.retType # NIL  THEN t.retType := update(a, t.retType) END;
    IF t.receiver # NIL THEN t.receiver := update(a, t.receiver) END;
    FOR i := 0 TO t.formals.tide - 1 DO
      f := t.formals.a[i];
      f.type := update(a, f.type);
    END;
  END TypeFix;

(* ============================================================ *)
(*  A type is "forced", i.e. must have its type			*)
(*  structure emitted to the symbol file if it is any of ...	*)
(*	i   : a local type with an exported TypId,		*)
(*	ii  : an imported type with an exported local alias,	*)
(*	iii : the base-type of a forced record,			*)
(*      iv  : a type with value semantics.			*)
(*  Partly forced types have structure but not methods emitted. *)
(* ============================================================ *)

  PROCEDURE MarkModule(ty : Sy.Type);
  BEGIN
    IF (ty.idnt # NIL) & (ty.idnt.dfScp # NIL) THEN
      INCL(ty.idnt.dfScp(Id.BlkId).xAttr, Sy.need);
    END;
  END MarkModule;

(* ---------------------------- *)

  PROCEDURE (i : Base)ConditionalMark*();
  BEGIN END ConditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Enum)ConditionalMark*();
  BEGIN END ConditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Opaque)ConditionalMark*();
  BEGIN
    MarkModule(i);
  END ConditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Pointer)ConditionalMark*();
  BEGIN
    IF i.force = Sy.noEmit THEN
      IF ~i.isImportedType() THEN
        i.force := Sy.forced;
        i.boundTp.ConditionalMark();
      ELSE
        MarkModule(i);
      END;
    END;
  END ConditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Record)ConditionalMark*();
    VAR idx : INTEGER;
	fTp : Sy.Type;
  BEGIN
    IF (i.force = Sy.noEmit) THEN
      IF i.isImportedType() THEN 
	i.force := Sy.partEmit;
        INCL(i.idnt.dfScp(Id.BlkId).xAttr, Sy.need);
        IF i.bindTp # NIL THEN i.bindTp.ConditionalMark() END;
        IF (i.baseTp # NIL) & 
	   ~(i.baseTp IS Base) THEN i.baseTp.ConditionalMark() END;
      ELSE 
	i.force := Sy.forced;
        IF i.bindTp # NIL THEN i.bindTp.UnconditionalMark() END;
        IF (i.baseTp # NIL) & 
	   ~(i.baseTp IS Base) THEN i.baseTp.UnconditionalMark() END;
      END;
      FOR idx := 0 TO i.fields.tide-1 DO
	fTp := i.fields.a[idx].type;
	fTp.ConditionalMark();
      END;
    END;
  END ConditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Array)ConditionalMark*();
  BEGIN
    IF (i.force = Sy.noEmit) THEN
      IF i.isImportedType() THEN 
        INCL(i.idnt.dfScp(Id.BlkId).xAttr, Sy.need);
      ELSE 
        i.force := Sy.forced;
        i.elemTp.ConditionalMark();
      END;
    END;
  END ConditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Procedure)ConditionalMark*();
  BEGIN
  END ConditionalMark;

(* ============================================================ *)
(*   Rules for unconditional marking don't care about imports.	*)
(* ============================================================ *)

  PROCEDURE (i : Base)UnconditionalMark*();
  BEGIN END UnconditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Opaque)UnconditionalMark*();
  BEGIN
    MarkModule(i);
  END UnconditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Enum)UnconditionalMark*();
  BEGIN
    MarkModule(i);
  END UnconditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Pointer)UnconditionalMark*();
  BEGIN
    IF (i.force # Sy.forced) THEN
      i.force := Sy.forced;
      i.boundTp.ConditionalMark();
      MarkModule(i);
    END;
  END UnconditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Record)UnconditionalMark*();
    VAR idx : INTEGER;
	fTp : Sy.Type;
  BEGIN
    IF (i.force # Sy.forced) THEN
      i.force := Sy.forced;
      IF i.baseTp # NIL THEN i.baseTp.UnconditionalMark() END;
      IF i.bindTp # NIL THEN i.bindTp.UnconditionalMark() END;
      FOR idx := 0 TO i.fields.tide-1 DO
	fTp := i.fields.a[idx].type;
	fTp.ConditionalMark();
      END;
      MarkModule(i);
    END;
  END UnconditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Array)UnconditionalMark*();
  BEGIN
    IF (i.force # Sy.forced) THEN
      i.force := Sy.forced;
      i.elemTp.ConditionalMark();
      MarkModule(i);
    END;
  END UnconditionalMark;

(* ---------------------------- *)

  PROCEDURE (i : Procedure)UnconditionalMark*();
  BEGIN
  END UnconditionalMark;

(* ============================================================ *)

  PROCEDURE (i : Pointer)boundRecTp*() : Sy.Type;
  BEGIN
    IF i.boundTp = NIL THEN RETURN NIL ELSE RETURN i.boundTp.boundRecTp() END;
  END boundRecTp;

(* ---------------------------- *)

  PROCEDURE (i : Record)boundRecTp*() : Sy.Type;
  BEGIN 
    RETURN i;
  END boundRecTp;

(* ---------------------------- *)

  PROCEDURE (i : Opaque)boundRecTp*() : Sy.Type;
  BEGIN 
    IF (i.resolved = NIL) OR
       (i.resolved IS Opaque) THEN 
      RETURN NIL;
    ELSE 
      RETURN i.resolved.boundRecTp();
    END;
  END boundRecTp;

(* ---------------------------- *)

  PROCEDURE (rec : Record)InsertMethod*(m : Sy.Idnt);
    VAR fwd : Sy.Idnt;
	mth : Id.MthId;
  BEGIN
    mth := m(Id.MthId);
    IF ~rec.symTb.enter(m.hash, m) THEN (* refused *)
      fwd := rec.symTb.lookup(m.hash);
      IF fwd.kind = Id.fwdMth THEN
	mth.CheckElab(fwd); 
	rec.symTb.Overwrite(m.hash, m);
      ELSIF fwd.kind = Id.fwdPrc THEN
	fwd.IdError(63);
      ELSIF ~fwd.isImported() THEN	(* why is this so? *)
	m.IdError(4);
      END;
    END;
    Sy.AppendIdnt(rec.methods, m);
  END InsertMethod;

(* ---------------------------- *)

  PROCEDURE (i : Procedure)boundRecTp*() : Sy.Type;
  BEGIN 
    IF i.receiver = NIL THEN RETURN NIL ELSE RETURN i.receiver.boundRecTp() END
  END boundRecTp;

(* ============================================================ *)

  PROCEDURE (i : Record)inheritedFeature*(hash : INTEGER) : Sy.Idnt;
  BEGIN
    IF (i.baseTp # NIL) & (i.baseTp IS Record) THEN 
      RETURN i.baseTp(Record).bindField(hash);
    ELSE 
      RETURN NIL;
    END;
  END inheritedFeature;

(* ============================================================ *)
(* 			Diagnostic methods			*)
(* ============================================================ *)

  PROCEDURE (s : Base)Diagnose*(i : INTEGER);
  BEGIN
    s.SuperDiag(i);
  END Diagnose;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Opaque)Diagnose*(i : INTEGER);
    VAR name : Lv.CharOpen;
  BEGIN
    s.SuperDiag(i);
    IF s.resolved # NIL THEN
      name := s.resolved.name();
      H.Indent(i+2); Con.WriteString("alias of " + name^); 
      s.resolved.SuperDiag(i+2);
    END;
  END Diagnose;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Array)Diagnose*(i : INTEGER);
  BEGIN
    s.SuperDiag(i);
    H.Indent(i+2); Con.WriteString("Element type");
    IF s.elemTp # NIL THEN
      Con.WriteLn; 
      s.elemTp.Diagnose(i+2);
    ELSE
      Con.WriteString(" NIL"); Con.WriteLn;
    END;
  END Diagnose;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Record)Diagnose*(i : INTEGER);
    VAR ix : INTEGER;
	id : Sy.Idnt;
        nm : FileNames.NameString;
  BEGIN
    s.SuperDiag(i);
    CASE s.recAtt OF
    | isAbs :	Console.WriteString(" ABSTRACT");	Console.WriteLn;
    | limit :	Console.WriteString(" LIMITED");	Console.WriteLn;
    | extns :	Console.WriteString(" EXTENSIBLE");	Console.WriteLn;
    | iFace :	Console.WriteString(" INTERFACE");	Console.WriteLn;
    ELSE
    END;
    IF    Sy.csInf IN s.xAttr THEN
      Console.WriteString(" [vos-interface]"); Console.WriteLn;
    ELSIF Sy.isCs  IN s.xAttr THEN
      Console.WriteString(" [vos-class]"); Console.WriteLn;
    ELSIF Sy.jvInf IN s.xAttr THEN
      Console.WriteString(" [java-interface]"); Console.WriteLn;
    ELSIF Sy.isJv  IN s.xAttr THEN
      Console.WriteString(" [java-class]"); Console.WriteLn;
    END; 
    H.Indent(i); Console.WriteString("fields"); Console.WriteLn;
    FOR ix := 0 TO s.fields.tide-1 DO 
      id := s.fields.a[ix];
      IF id # NIL THEN id.Diagnose(i+4) END;
    END;
    IF G.verbose THEN
      H.Indent(i); Console.WriteString("methods"); Console.WriteLn;
      FOR ix := 0 TO s.methods.tide-1 DO 
        id := s.methods.a[ix];
        IF id # NIL THEN id.Diagnose(i+4) END;
      END;
      H.Indent(i); Console.WriteString("names"); Console.WriteLn;
      s.symTb.Dump(i+4);
    END;
    IF s.baseTp # NIL THEN
      H.Indent(i); Console.WriteString("base type"); Console.WriteLn;
      s.baseTp.Diagnose(i+4);
    END;
    Sy.DoXName(i, s.xName);
  END Diagnose;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Enum)Diagnose*(i : INTEGER);
    VAR ix : INTEGER;
	id : Sy.Idnt;
        nm : FileNames.NameString;
  BEGIN
    s.SuperDiag(i);
    H.Indent(i); Console.WriteString("consts"); Console.WriteLn;
    FOR ix := 0 TO s.statics.tide-1 DO 
      id := s.statics.a[ix];
      IF id # NIL THEN id.Diagnose(i+4) END;
    END;
    Sy.DoXName(i, s.xName);
  END Diagnose;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Pointer)Diagnose*(i : INTEGER);
  BEGIN
    s.SuperDiag(i);
    H.Indent(i+2); Con.WriteString("Bound type");
    IF s.boundTp # NIL THEN
      Con.WriteLn; 
      s.boundTp.Diagnose(i+2);
    ELSE
      Con.WriteString(" NIL"); Con.WriteLn;
    END;
  END Diagnose;

(* ---------------------------------------------------- *)
  PROCEDURE^ qualname(id : Sy.Idnt) : Lv.CharOpen;
(* ---------------------------------------------------- *)

  PROCEDURE (s : Procedure)DiagFormalType*(i : INTEGER);
    VAR ix : INTEGER;
	nm : FileNames.NameString;
  BEGIN
    IF s.formals.tide = 0 THEN
      Console.WriteString("()");
    ELSE
      Console.Write("(");
      Console.WriteLn;
      FOR ix := 0 TO s.formals.tide-1 DO
	H.Indent(i+4);
        s.formals.a[ix].DiagPar();
        IF ix < s.formals.tide-1 THEN Console.Write(";"); Console.WriteLn END;
      END;
      Console.Write(")");
    END;
    IF s.retType # NIL THEN
      Console.WriteString(" : ");
      Console.WriteString(qualname(s.retType.idnt));
    END;
  END DiagFormalType;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Procedure)Diagnose*(i : INTEGER);
    VAR ix : INTEGER;
  BEGIN
    H.Indent(i);
    IF s.receiver # NIL THEN
      Console.Write("(");
      Console.WriteString(s.name());
      Console.Write(")");
    END;
    Console.WriteString("PROC");
    s.DiagFormalType(i+4);
    Console.WriteLn;
    Sy.DoXName(i, s.xName);
  END Diagnose;

(* ---------------------------------------------------- *)
(* ---------------------------------------------------- *)

  PROCEDURE qualname(id : Sy.Idnt) : Lv.CharOpen;
  BEGIN
    IF id = NIL THEN
      RETURN nilStr;
    ELSIF (id.dfScp = NIL) OR (id.dfScp.kind = Id.modId) THEN
      RETURN Sy.getName.ChPtr(id);
    ELSE
      RETURN Lv.strToCharOpen
		(Sy.getName.ChPtr(id.dfScp)^ + "." + Sy.getName.ChPtr(id)^);
    END;
  END qualname;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Base)name*() : Lv.CharOpen;
  BEGIN
    IF s.idnt = NIL THEN
      RETURN Lv.strToCharOpen("Anon-base-type");
    ELSE
      RETURN Sy.getName.ChPtr(s.idnt);
    END;
  END name;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Enum)name*() : Lv.CharOpen;
  BEGIN
    IF s.idnt = NIL THEN
      RETURN Lv.strToCharOpen("Anon-enum-type");
    ELSE
      RETURN Sy.getName.ChPtr(s.idnt);
    END;
  END name;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Opaque)name*() : Lv.CharOpen;
  BEGIN
    IF s.idnt = NIL THEN
      IF s.kind = namTp THEN
        RETURN Lv.strToCharOpen("Anon-opaque");
      ELSE
        RETURN Lv.strToCharOpen("Anon-temporary");
      END;
    ELSE
      RETURN qualname(s.idnt);
    END;
  END name;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Array)name*() : Lv.CharOpen;
    VAR elNm : Lv.CharOpen;
  BEGIN
    IF s.idnt = NIL THEN
      IF s.elemTp = NIL THEN elNm := nilStr ELSE elNm := s.elemTp.name() END;
      RETURN Lv.strToCharOpen("ARRAY OF " + elNm^);
    ELSE
      RETURN qualname(s.idnt);
    END;
  END name;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Record)name*() : Lv.CharOpen;
  BEGIN
    IF s.bindTp # NIL THEN 
      RETURN Lv.strToCharOpen(s.bindTp.name()^ + "^");
    ELSIF s.idnt = NIL THEN
      RETURN Lv.strToCharOpen("Anon-record");
    ELSE
      RETURN qualname(s.idnt);
    END;
  END name;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Pointer)name*() : Lv.CharOpen;
    VAR elNm : Lv.CharOpen;
  BEGIN
    IF s.idnt = NIL THEN
      IF s.boundTp = NIL THEN elNm := nilStr ELSE elNm := s.boundTp.name() END;
      RETURN Lv.strToCharOpen("POINTER TO " + elNm^);
    ELSE
      RETURN qualname(s.idnt);
    END;
  END name;

(* ---------------------------------------------------- *)

  PROCEDURE (s : Procedure)name*() : Lv.CharOpen;
  BEGIN
    IF s.idnt = NIL THEN
      RETURN Lv.strToCharOpen("Anon-opaque-type");
    ELSE
      RETURN qualname(s.idnt);
    END;
  END name;

(* ============================================================ *)
BEGIN (* ====================================================== *)
  NEW(anyRecTp);
  nilStr := Lv.strToCharOpen("<nil>");
END TypeDesc. (* ============================================== *)
(* ============================================================ *)
