(* ==================================================================== *)
(*									*)
(*  Visitor pattern for the Gardens Point Component Pascal Compiler.	*)
(*  This module defines various extensions of type Symbols.SymForAll	*)
(*	Copyright (c) John Gough 1999, 2000.				*)
(*									*)
(* ==================================================================== *)

MODULE Visitor;

  IMPORT 
	GPCPcopyright,
	Console,
	Symbols,
	IdDesc,
	CPascalS,
	LitValue,
	T := TypeDesc,
	NameHash;

(* ============================================================ *)

  TYPE
    Resolver* = POINTER TO RECORD (Symbols.SymForAll) END;

(* -------------------------------------------- *)

  TYPE
    ImplementedCheck* = POINTER TO RECORD (Symbols.SymForAll) END;

(* -------------------------------------------- *)

  TYPE
    Accumulator* = POINTER TO RECORD (Symbols.SymForAll)
		     missing  : Symbols.SymbolTable;
		     recordTb : TypeDesc.Record;
		     checkCov : BOOLEAN;
		   END;

(* ============================================================ *)

  PROCEDURE newAccumulator(target : TypeDesc.Record) : Accumulator;
  (** Create a new symbol table and add to it all of the abstract *)
  (*  methods which are not concrete in the target scope space.   *)
    VAR tmp : Accumulator;
  BEGIN 
    NEW(tmp); 
    tmp.recordTb := target;
    tmp.checkCov := FALSE;	(* not an interface *)
    RETURN tmp;
  END newAccumulator;

 (* --------------------------- *)

  PROCEDURE newInterfaceCheck(target : TypeDesc.Record) : Accumulator;
  (** Create a new symbol table and add to it all of the abstract *)
  (*  methods which are not concrete in the target scope space.   *)
    VAR tmp : Accumulator;
  BEGIN 
    NEW(tmp); 
    tmp.recordTb := target;
    tmp.checkCov := TRUE;	(* is an interface *)
    RETURN tmp;
  END newInterfaceCheck;

 (* --------------------------- *)

  PROCEDURE (sfa : Accumulator)Op*(id : Symbols.Idnt);
    VAR anyId : Symbols.Idnt;
	pType : T.Procedure;
	junk  : BOOLEAN;
  BEGIN
    IF id.isAbstract() THEN
      anyId := sfa.recordTb.bindField(id.hash);
      IF (anyId = NIL) OR anyId.isAbstract() THEN
	junk := sfa.missing.enter(id.hash, id);
      ELSIF sfa.checkCov & (anyId.kind = IdDesc.conMth) THEN
	pType := id.type(T.Procedure);
	pType.CheckCovariance(anyId);
	IF id.vMod # anyId.vMod THEN anyId.IdError(163) END;
      END;
    END;
  END Op;

(* -------------------------------------------- *)

  PROCEDURE newImplementedCheck*() : ImplementedCheck;
    VAR tmp : ImplementedCheck;
  BEGIN NEW(tmp); RETURN tmp END newImplementedCheck;

 (* --------------------------- *)

  PROCEDURE (sfa : ImplementedCheck)Op*(id : Symbols.Idnt);
    VAR acc : Accumulator;
	rTp : T.Record;
	nTp : T.Record;
	bTp : Symbols.Type;
	idx : INTEGER;
  BEGIN
    IF id.kind = IdDesc.typId THEN
      rTp := id.type.boundRecTp()(T.Record);
      IF (rTp # NIL) &			(* ==> this is a record type	*)
	 ~rTp.isAbsRecType() &		(* ==> this rec is NOT abstract *)
	 (rTp.baseTp # NIL) THEN	(* ==> this extends some type   *)
	bTp := rTp.baseTp;
	IF bTp.isAbsRecType() THEN	(* ==> and base _is_ abstract.	*)
         (* 
	  *  This is a concrete record extending an abstract record.
	  *  By now, all inherited abstract methods must have been
	  *  resolved to concrete methods.  Traverse up the base
	  *  hierarchy accumulating unimplemented methods.
	  *)
	  acc := newAccumulator(rTp);
	  REPEAT
	    nTp := bTp(T.Record);	(* guaranteed for first time	*)
	    bTp := nTp.baseTp;
	    nTp.symTb.Apply(acc);
	  UNTIL (bTp = NIL) OR bTp.isBaseType();
         (*
	  *  Now we turn the missing table into a list.
	  *)
	  IF ~acc.missing.isEmpty() THEN
            CPascalS.SemError.RepSt1(121, Symbols.dumpList(acc.missing),
						id.token.lin, id.token.col);
	  END;
        END;
	IF rTp.interfaces.tide > 0 THEN
	 (*
	  *   The record rTp claims to implement interfaces.
	  *   We must check conformance to the contract.
	  *)
	  acc := newInterfaceCheck(rTp);
	  FOR idx := 0 TO rTp.interfaces.tide-1 DO
	    bTp := rTp.interfaces.a[idx].type.boundRecTp();
	    bTp(T.Record).symTb.Apply(acc);
	  END;
         (*
	  *  Now we turn the missing table into a list.
	  *)
	  IF ~acc.missing.isEmpty() THEN
            CPascalS.SemError.RepSt1(159, Symbols.dumpList(acc.missing),
						id.token.lin, id.token.col);
	  END;
        END;
      END;
    END;
  END Op;

(* -------------------------------------------- *)

  PROCEDURE newResolver*() : Resolver;
    VAR tmp : Resolver;
  BEGIN 
    NEW(tmp); 
    RETURN tmp;
  END newResolver;

 (* --------------------------- *)

  PROCEDURE (sfa : Resolver)Op*(id : Symbols.Idnt);
    VAR idTp : Symbols.Type;
  BEGIN
    IF (id.kind = IdDesc.typId) OR (id.kind = IdDesc.varId) THEN
      idTp := id.type;
      IF idTp # NIL THEN 
	idTp := idTp.resolve(1);
	IF idTp # NIL THEN
	  WITH idTp : T.Array DO
	     IF idTp.isOpenArrType() THEN id.IdError(67) END;
	  | idTp : T.Record DO
	     IF id.kind = IdDesc.varId THEN idTp.InstantiateCheck(id.token) END;
	  ELSE
	  END;
	END;
      END;
      id.type := idTp;
    END;
  END Op;

(* ============================================================ *)
END Visitor.  (* ============================================== *)
(* ============================================================ *)

